@@ -2,7 +2,7 @@ use crate::image_view::BstImageView;
 use crate::{misc, Basalt};
 use crossbeam::deque::{Injector, Steal};
 use crossbeam::sync::{Parker, Unparker};
-use ilmenite::ImtWeight;
+use ilmenite::{ImtImageView, ImtWeight};
 use image::{self, GenericImageView};
 use ordered_float::OrderedFloat;
 use parking_lot::{Condvar, Mutex};
@@ -10,20 +10,20 @@ use std::collections::HashMap;
 use std::fs::File;
 use std::io::Read;
 use std::path::PathBuf;
-use std::sync::atomic::{self, AtomicBool};
 use std::sync::Arc;
-use std::thread;
 use std::time::Instant;
+use std::{fmt, thread};
 use vulkano::buffer::cpu_access::CpuAccessibleBuffer;
 use vulkano::buffer::{BufferAccess, BufferUsage as VkBufferUsage};
 use vulkano::command_buffer::{
 	AutoCommandBufferBuilder, CommandBufferUsage, PrimaryAutoCommandBuffer,
 	PrimaryCommandBuffer,
 };
+use vulkano::format::NumericType as VkFormatType;
 use vulkano::image::immutable::ImmutableImage;
 use vulkano::image::{
-	ImageCreateFlags, ImageDimensions as VkImgDimensions, ImageUsage as VkImageUsage,
-	MipmapsCount, StorageImage,
+	ImageAccess, ImageCreateFlags, ImageDimensions as VkImgDimensions,
+	ImageUsage as VkImageUsage, MipmapsCount, SampleCount, StorageImage,
 };
 use vulkano::sampler::Sampler;
 use vulkano::sync::GpuFuture;
@@ -31,37 +31,48 @@ use vulkano::sync::GpuFuture;
 const PRINT_UPDATE_TIME: bool = false;
 
 #[inline]
-fn srgb_to_linear_d8(v: u8) -> u8 {
-	let mut f = v as f32 / 255.0;
-
+fn srgb_to_linear(mut f: f32) -> u16 {
 	if f < 0.04045 {
 		f /= 12.92;
 	} else {
 		f = ((f + 0.055) / 1.055).powf(2.4)
 	}
 
-	f = (f * 255.0).round();
+	f = (f * u16::max_value() as f32).round();
 
-	if f > 255.0 {
-		f = 255.0;
+	if f > u16::max_value() as f32 {
+		f = u16::max_value() as f32;
 	} else if f < 0.0 {
 		f = 0.0;
 	}
 
-	f as u8
+	f as u16
 }
 
 #[inline]
-fn linear_to_srgb(v: u8) -> u8 {
-	let mut v = ((((v as f32 / 255.0).powf(1.0 / 2.4) * 1.005) - 0.055) * 255.0).round();
+fn linear_to_srgb(v: f32) -> u16 {
+	let mut v = (((v.powf(1.0 / 2.4) * 1.005) - 0.055) * u16::max_value() as f32).round();
 
-	if v > 255.0 {
-		v = 255.0;
+	if v > u16::max_value() as f32 {
+		v = u16::max_value() as f32;
 	} else if v < 0.0 {
 		v = 0.0;
 	}
 
-	v as u8
+	v as u16
+}
+
+#[inline]
+fn f32_to_u16(mut v: f32) -> u16 {
+	v *= u16::max_value() as f32;
+
+	if v > u16::max_value() as f32 {
+		v = u16::max_value() as f32;
+	} else if v < 0.0 {
+		v = 0.0;
+	}
+
+	v.trunc() as u16
 }
 
 const CELL_WIDTH: u32 = 32;
@@ -133,11 +144,23 @@ pub struct ImageDims {
 	pub h: u32,
 }
 
-#[derive(Debug, Clone)]
+#[derive(Clone)]
 pub enum ImageData {
 	D8(Vec<u8>),
-	#[doc(hidden)]
-	__Nonexhaustive,
+	D16(Vec<u16>),
+	Imt(Arc<ImtImageView>),
+	Bst(Arc<BstImageView>),
+}
+
+impl fmt::Debug for ImageData {
+	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+		match self {
+			ImageData::D8(_) => write!(f, "ImageData::D8"),
+			ImageData::D16(_) => write!(f, "ImageData::D16"),
+			ImageData::Imt(_) => write!(f, "ImageData::Imt"),
+			ImageData::Bst(_) => write!(f, "ImageData::Bst"),
+		}
+	}
 }
 
 #[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
@@ -148,8 +171,8 @@ pub enum ImageType {
 	SRGBA,
 	SRGB,
 	SMono,
-	Glyph,
 	YUV444,
+	Raw,
 }
 
 impl ImageType {
@@ -161,34 +184,61 @@ impl ImageType {
 			&ImageType::SRGBA => 4,
 			&ImageType::SRGB => 3,
 			&ImageType::SMono => 1,
-			&ImageType::Glyph => 1,
 			&ImageType::YUV444 => 3,
+			&ImageType::Raw => 0,
 		}
 	}
 }
 
+#[derive(Debug, Clone)]
 pub struct Image {
 	ty: ImageType,
 	dims: ImageDims,
 	data: ImageData,
 }
 
+fn image_atlas_compatible(img: &dyn ImageAccess) -> Result<(), String> {
+	if img.samples() != SampleCount::Sample1 {
+		return Err(String::from("Source image must not be multisampled. "));
+	}
+
+	match img.format().type_color() {
+		Some(color_type) =>
+			match color_type {
+				VkFormatType::UNORM => (),
+				_ => return Err(format!("Source image must be an unorm numeric type.")),
+			},
+		None => return Err(format!("Source image must be of a color format.")),
+	}
+
+	Ok(())
+}
+
 impl Image {
-	pub fn new(ty: ImageType, dims: ImageDims, mut data: ImageData) -> Result<Image, String> {
+	pub fn new(ty: ImageType, dims: ImageDims, data: ImageData) -> Result<Image, String> {
+		if ty == ImageType::Raw {
+			return Err(format!(
+				"This method can not create raw images. Use `from_imt` or `from_bst`."
+			));
+		}
+
 		let expected_len = dims.w as usize * dims.h as usize * ty.components();
 
 		if expected_len == 0 {
-			return Err(format!("Image can't be empty"));
+			return Err(format!("Image can't be empty."));
 		}
 
-		match &mut data {
-			&mut ImageData::D8(ref mut d) =>
-				if d.len() > expected_len {
-					d.truncate(expected_len);
-				} else if d.len() < expected_len {
-					return Err(format!("Data length doesn't match the provided dimensions."));
-				},
-			_ => unreachable!(),
+		let actual_len = match &data {
+			ImageData::D8(d) => d.len(),
+			ImageData::D16(d) => d.len(),
+			_ => return Err(format!("`Image::new()` can only create D8 & D16 images.")),
+		};
+
+		if actual_len != expected_len {
+			return Err(format!(
+				"Data len doesn't match the provided dimensions! {} != {}",
+				actual_len, expected_len
+			));
 		}
 
 		Ok(Image {
@@ -198,167 +248,284 @@ impl Image {
 		})
 	}
 
+	pub fn from_imt(imt: Arc<ImtImageView>) -> Result<Image, String> {
+		let dims = match imt.dimensions() {
+			VkImgDimensions::Dim2d {
+				width,
+				height,
+				array_layers,
+			} => {
+				if array_layers != 1 {
+					return Err(format!("array_layers != 1"));
+				}
+
+				ImageDims {
+					w: width,
+					h: height,
+				}
+			},
+			_ => {
+				return Err(format!("Only 2d images are supported."));
+			},
+		};
+
+		image_atlas_compatible(&imt)?;
+
+		Ok(Image {
+			ty: ImageType::Raw,
+			dims,
+			data: ImageData::Imt(imt),
+		})
+	}
+
+	pub fn from_bst(bst: Arc<BstImageView>) -> Result<Image, String> {
+		let dims = match bst.dimensions() {
+			VkImgDimensions::Dim2d {
+				width,
+				height,
+				array_layers,
+			} => {
+				if array_layers != 1 {
+					return Err(format!("array_layers != 1"));
+				}
+
+				ImageDims {
+					w: width,
+					h: height,
+				}
+			},
+			_ => {
+				return Err(format!("Only 2d images are supported."));
+			},
+		};
+
+		image_atlas_compatible(&bst)?;
+
+		Ok(Image {
+			ty: ImageType::Raw,
+			dims,
+			data: ImageData::Bst(bst),
+		})
+	}
+
 	pub fn into_data(self) -> ImageData {
 		self.data
 	}
 
-	pub fn to_srgba(self) -> Self {
-		if let ImageData::D8(data) = self.data {
-			let mut srgba = Vec::with_capacity(data.len() / self.ty.components() * 4);
+	pub fn to_16b_srgba(self) -> Self {
+		let (len, mut data) = match self.data {
+			ImageData::D8(data) =>
+				(
+					data.len() / self.ty.components(),
+					Box::new(data.into_iter().map(|v| v as f32 / u8::max_value() as f32))
+						as Box<dyn Iterator<Item = f32>>,
+				),
+			ImageData::D16(data) =>
+				(
+					data.len() / self.ty.components(),
+					Box::new(data.into_iter().map(|v| v as f32 / u16::max_value() as f32))
+						as Box<dyn Iterator<Item = f32>>,
+				),
+			_ => return self,
+		};
 
-			match self.ty {
-				ImageType::LRGBA =>
-					for v in data {
-						srgba.push(linear_to_srgb(v));
-					},
-				ImageType::LRGB =>
-					for v in data {
-						srgba.push(linear_to_srgb(v));
+		let srgba: Vec<u16> = match self.ty {
+			ImageType::LRGBA => data.map(|v| linear_to_srgb(v)).collect(),
+			ImageType::LRGB => {
+				let mut srgba: Vec<u16> = Vec::with_capacity(len * 4);
 
-						if srgba.len() % 4 == 2 {
-							srgba.push(255);
-						}
-					},
-				ImageType::LMono =>
-					for mut v in data {
-						v = linear_to_srgb(v);
-						srgba.push(v);
-						srgba.push(v);
-						srgba.push(v);
-						srgba.push(255);
-					},
-				ImageType::SMono =>
-					for v in data {
-						srgba.push(v);
-						srgba.push(v);
-						srgba.push(v);
-						srgba.push(255);
-					},
-				ImageType::SRGBA => srgba = data,
-				ImageType::SRGB =>
-					for v in data {
-						srgba.push(v);
+				for v in data {
+					srgba.push(linear_to_srgb(v));
 
-						if srgba.len() % 4 == 2 {
-							srgba.push(255);
-						}
-					},
-				ImageType::Glyph =>
-					for v in data {
-						srgba.push(0);
-						srgba.push(0);
-						srgba.push(0);
-						srgba.push(v);
-					},
-				ImageType::YUV444 =>
-					for chunk in data.chunks_exact(3) {
-						let components = [
-							chunk[0] as f32 + (1.402 * (chunk[2] as f32 - 128.0)),
-							chunk[0] as f32 + (0.344 * (chunk[1] as f32 - 128.0))
-								- (0.714 * (chunk[2] as f32 - 128.0)),
-							chunk[0] as f32 + (1.772 * (chunk[1] as f32 - 128.0)),
-						];
+					if srgba.len() % 4 == 2 {
+						srgba.push(u16::max_value());
+					}
+				}
 
-						for v in &components {
-							srgba.push(*v as u8);
-						}
+				srgba
+			},
+			ImageType::LMono => {
+				let mut srgba: Vec<u16> = Vec::with_capacity(len * 4);
 
-						srgba.push(255);
-					},
-			}
+				for v in data {
+					let v = linear_to_srgb(v);
+					srgba.push(v);
+					srgba.push(v);
+					srgba.push(v);
+					srgba.push(u16::max_value());
+				}
 
-			Image {
-				ty: ImageType::SRGBA,
-				dims: self.dims,
-				data: ImageData::D8(srgba),
-			}
-		} else {
-			unreachable!()
+				srgba
+			},
+			ImageType::SMono => {
+				let mut srgba: Vec<u16> = Vec::with_capacity(len * 4);
+
+				for v in data {
+					let v = f32_to_u16(v);
+					srgba.push(v);
+					srgba.push(v);
+					srgba.push(v);
+					srgba.push(u16::max_value());
+				}
+
+				srgba
+			},
+			ImageType::SRGBA => data.map(|v| f32_to_u16(v)).collect(),
+			ImageType::SRGB => {
+				let mut srgba: Vec<u16> = Vec::with_capacity(len * 4);
+
+				for v in data {
+					srgba.push(f32_to_u16(v));
+
+					if srgba.len() % 4 == 2 {
+						srgba.push(u16::max_value());
+					}
+				}
+
+				srgba
+			},
+			ImageType::YUV444 => {
+				let mut srgba: Vec<u16> = Vec::with_capacity(len * 4);
+
+				loop {
+					let y = match data.next() {
+						Some(some) => some,
+						None => break,
+					};
+
+					let u = data.next().unwrap();
+					let v = data.next().unwrap();
+
+					let components = [
+						y + (1.402 * (v - 0.5)),
+						y + (0.344 * (u - 0.5)) - (0.714 * (v - 0.5)),
+						y + (1.772 * (u - 0.5)),
+					];
+
+					for v in &components {
+						srgba.push(f32_to_u16(*v));
+					}
+
+					srgba.push(u16::max_value());
+				}
+
+				srgba
+			},
+			_ => panic!("Unexpected ImageType"),
+		};
+
+		Image {
+			ty: ImageType::SRGBA,
+			dims: self.dims,
+			data: ImageData::D16(srgba),
 		}
 	}
 
-	pub fn to_lrgba(self) -> Self {
-		if let ImageData::D8(data) = self.data {
-			let mut lrgba = Vec::with_capacity(data.len() / self.ty.components() * 4);
+	pub fn to_16b_lrgba(self) -> Self {
+		let (len, mut data) = match self.data {
+			ImageData::D8(data) =>
+				(
+					data.len() / self.ty.components(),
+					Box::new(data.into_iter().map(|v| v as f32 / u8::max_value() as f32))
+						as Box<dyn Iterator<Item = f32>>,
+				),
+			ImageData::D16(data) =>
+				(
+					data.len() / self.ty.components(),
+					Box::new(data.into_iter().map(|v| v as f32 / u16::max_value() as f32))
+						as Box<dyn Iterator<Item = f32>>,
+				),
+			_ => return self,
+		};
 
-			match self.ty {
-				ImageType::LRGBA => lrgba = data,
-				ImageType::LRGB =>
-					for v in data {
-						lrgba.push(v);
+		let lrgba: Vec<u16> = match self.ty {
+			ImageType::LRGBA => data.map(|v| f32_to_u16(v)).collect(),
+			ImageType::LRGB => {
+				let mut lrgba: Vec<u16> = Vec::with_capacity(len * 4);
 
-						if lrgba.len() % 4 == 2 {
-							lrgba.push(255);
-						}
-					},
-				ImageType::LMono =>
-					for v in data {
-						lrgba.push(v);
-						lrgba.push(v);
-						lrgba.push(v);
-						lrgba.push(255);
-					},
-				ImageType::SMono =>
-					for mut v in data {
-						v = srgb_to_linear_d8(v);
-						lrgba.push(v);
-						lrgba.push(v);
-						lrgba.push(v);
-						lrgba.push(255);
-					},
-				ImageType::SRGBA =>
-					for v in data {
-						lrgba.push(srgb_to_linear_d8(v));
-					},
-				ImageType::SRGB =>
-					for v in data {
-						lrgba.push(srgb_to_linear_d8(v));
+				for v in data {
+					lrgba.push(f32_to_u16(v));
 
-						if lrgba.len() % 4 == 2 {
-							lrgba.push(255);
-						}
-					},
-				ImageType::Glyph =>
-					for v in data {
-						lrgba.push(0);
-						lrgba.push(0);
-						lrgba.push(0);
-						lrgba.push(srgb_to_linear_d8(v));
-					},
-				ImageType::YUV444 =>
-					for chunk in data.chunks_exact(3) {
-						let mut components = [
-							chunk[0] as f32 + (1.402 * (chunk[2] as f32 - 128.0)),
-							chunk[0] as f32 + (0.344 * (chunk[1] as f32 - 128.0))
-								- (0.714 * (chunk[2] as f32 - 128.0)),
-							chunk[0] as f32 + (1.772 * (chunk[1] as f32 - 128.0)),
-						];
+					if lrgba.len() % 4 == 2 {
+						lrgba.push(u16::max_value());
+					}
+				}
 
-						for v in &mut components {
-							*v = ((*v + (0.055 * 255.0)) / 1.055).powf(2.4).round();
+				lrgba
+			},
+			ImageType::LMono => {
+				let mut lrgba: Vec<u16> = Vec::with_capacity(len * 4);
 
-							if *v > 255.0 {
-								*v = 255.0;
-							} else if *v < 0.0 {
-								*v = 0.0;
-							}
-						}
+				for v in data {
+					lrgba.push(f32_to_u16(v));
+					lrgba.push(f32_to_u16(v));
+					lrgba.push(f32_to_u16(v));
+					lrgba.push(u16::max_value());
+				}
 
-						for v in &components {
-							lrgba.push(*v as u8);
-						}
+				lrgba
+			},
+			ImageType::SMono => {
+				let mut lrgba: Vec<u16> = Vec::with_capacity(len * 4);
 
-						lrgba.push(255);
-					},
-			}
+				for v in data {
+					let v = srgb_to_linear(v);
+					lrgba.push(v);
+					lrgba.push(v);
+					lrgba.push(v);
+					lrgba.push(u16::max_value());
+				}
 
-			Image {
-				ty: ImageType::LRGBA,
-				dims: self.dims,
-				data: ImageData::D8(lrgba),
-			}
-		} else {
-			unreachable!()
+				lrgba
+			},
+			ImageType::SRGBA => data.map(|v| srgb_to_linear(v)).collect(),
+			ImageType::SRGB => {
+				let mut lrgba: Vec<u16> = Vec::with_capacity(len * 4);
+
+				for v in data {
+					lrgba.push(srgb_to_linear(v));
+
+					if lrgba.len() % 4 == 2 {
+						lrgba.push(u16::max_value());
+					}
+				}
+
+				lrgba
+			},
+			ImageType::YUV444 => {
+				let mut lrgba: Vec<u16> = Vec::with_capacity(len * 4);
+
+				loop {
+					let y = match data.next() {
+						Some(some) => some,
+						None => break,
+					};
+
+					let u = data.next().unwrap();
+					let v = data.next().unwrap();
+
+					let mut components = [
+						y + (1.402 * (v - 0.5)),
+						y + (0.344 * (u - 0.5)) - (0.714 * (v - 0.5)),
+						y + (1.772 * (u - 0.5)),
+					];
+
+					for v in &mut components {
+						lrgba.push(srgb_to_linear(*v));
+					}
+
+					lrgba.push(u16::max_value());
+				}
+
+				lrgba
+			},
+			_ => panic!("Unexpected ImageType"),
+		};
+
+		Image {
+			ty: ImageType::LRGBA,
+			dims: self.dims,
+			data: ImageData::D16(lrgba),
 		}
 	}
 }
@@ -404,14 +571,15 @@ pub struct Atlas {
 	basalt: Arc<Basalt>,
 	cmd_queue: Injector<Command>,
 	empty_image: Arc<BstImageView>,
-	default_sampler: Arc<Sampler>,
+	linear_sampler: Arc<Sampler>,
+	nearest_sampler: Arc<Sampler>,
 	unparker: Unparker,
 	image_views: Mutex<Option<(Instant, Arc<HashMap<AtlasImageID, Arc<BstImageView>>>)>>,
 }
 
 impl Atlas {
 	pub fn new(basalt: Arc<Basalt>) -> Arc<Self> {
-		let default_sampler = Sampler::unnormalized(
+		let linear_sampler = Sampler::unnormalized(
 			basalt.device(),
 			vulkano::sampler::Filter::Linear,
 			vulkano::sampler::UnnormalizedSamplerAddressMode::ClampToBorder(
@@ -423,6 +591,18 @@ impl Atlas {
 		)
 		.unwrap();
 
+		let nearest_sampler = Sampler::unnormalized(
+			basalt.device(),
+			vulkano::sampler::Filter::Nearest,
+			vulkano::sampler::UnnormalizedSamplerAddressMode::ClampToBorder(
+				vulkano::sampler::BorderColor::IntTransparentBlack,
+			),
+			vulkano::sampler::UnnormalizedSamplerAddressMode::ClampToBorder(
+				vulkano::sampler::BorderColor::IntTransparentBlack,
+			),
+		)
+		.unwrap();
+
 		let empty_image = BstImageView::from_immutable(
 			ImmutableImage::from_iter(
 				vec![255, 255, 255, 255].into_iter(),
@@ -432,8 +612,8 @@ impl Atlas {
 					array_layers: 1,
 				},
 				MipmapsCount::One,
-				vulkano::format::Format::R8G8B8A8Srgb,
-				basalt.compute_queue.clone(), // TODO: Secondary graphics queue
+				basalt.formats_in_use().atlas,
+				basalt.secondary_graphics_queue().unwrap_or_else(|| basalt.graphics_queue()),
 			)
 			.unwrap()
 			.0,
@@ -446,7 +626,8 @@ impl Atlas {
 		let atlas_ret = Arc::new(Atlas {
 			basalt,
 			unparker,
-			default_sampler,
+			linear_sampler,
+			nearest_sampler,
 			empty_image,
 			cmd_queue: Injector::new(),
 			image_views: Mutex::new(None),
@@ -551,8 +732,11 @@ impl Atlas {
 
 				let mut cmd_buf = AutoCommandBufferBuilder::primary(
 					atlas.basalt.device(),
-					atlas.basalt.compute_queue_ref().family(), /* TODO: Secondary graphics
-					                                            * queue */
+					atlas
+						.basalt
+						.secondary_graphics_queue_ref()
+						.unwrap_or_else(|| atlas.basalt.graphics_queue_ref())
+						.family(),
 					CommandBufferUsage::OneTimeSubmit,
 				)
 				.unwrap();
@@ -571,15 +755,21 @@ impl Atlas {
 				}
 
 				if execute {
-					drop(
-						cmd_buf
-                            .build()
-                            .unwrap()
-                            .execute(atlas.basalt.compute_queue()) // TODO: Secondary graphics queue
-                            .unwrap()
-                            .then_signal_fence_and_flush()
-                            .unwrap(),
-					);
+					cmd_buf
+						.build()
+						.unwrap()
+						.execute(
+							atlas
+								.basalt
+								.secondary_graphics_queue()
+								.unwrap_or_else(|| atlas.basalt.graphics_queue()),
+						)
+						.unwrap()
+						.then_signal_fence_and_flush()
+						.unwrap()
+						.wait(None)
+						.unwrap();
+
 					let mut draw_map = HashMap::new();
 
 					for (i, atlas_image) in atlas_images.iter_mut().enumerate() {
@@ -626,10 +816,16 @@ impl Atlas {
 		self.empty_image.clone()
 	}
 
-	/// The recommended default sampler for when using atlas image views. May be used elsewhere,
-	/// but probably more ideal to create your own.
-	pub fn default_sampler(&self) -> Arc<Sampler> {
-		self.default_sampler.clone()
+	/// An unnormalized, linear filter, clamp to transparent black border `vulkano::Sampler`
+	/// primary used for sampling atlas images. May be useful outside of Basalt.
+	pub fn linear_sampler(&self) -> Arc<Sampler> {
+		self.linear_sampler.clone()
+	}
+
+	/// An unnormalized, nearest filter, clamp to transparent black border `vulkano::Sampler`
+	/// primary used for sampling atlas images. May be useful outside of Basalt.
+	pub fn nearest_sampler(&self) -> Arc<Sampler> {
+		self.nearest_sampler.clone()
 	}
 
 	/// Remove a sub image. Currently not implemented.
@@ -666,7 +862,7 @@ impl Atlas {
 		image: Image,
 	) -> Result<Coords, String> {
 		let response = CommandResponse::new();
-		self.cmd_queue.push(Command::Upload(response.clone(), cache_id, image.to_srgba()));
+		self.cmd_queue.push(Command::Upload(response.clone(), cache_id, image.to_16b_lrgba()));
 		self.unparker.unpark();
 		response.wait_for_response()
 	}
@@ -682,7 +878,7 @@ impl Atlas {
 		};
 
 		let (w, h, data) = match image::load_from_memory(bytes.as_slice()) {
-			Ok(image) => (image.width(), image.height(), image.to_rgba8().into_vec()),
+			Ok(image) => (image.width(), image.height(), image.to_rgba16().into_vec()),
 			Err(e) => return Err(format!("Failed to read image: {}", e)),
 		};
 
@@ -697,10 +893,10 @@ impl Atlas {
 				w,
 				h,
 			},
-			ImageData::D8(data),
+			ImageData::D16(data),
 		)
 		.map_err(|e| format!("Invalid Image: {}", e))?;
-		self.load_image(cache_id, image)
+		self.load_image(cache_id, image.to_16b_lrgba())
 	}
 
 	pub fn load_image_from_path<P: Into<PathBuf>>(&self, path: P) -> Result<Coords, String> {
@@ -779,7 +975,6 @@ struct AtlasImage {
 	update: Option<usize>,
 	sto_imgs: Vec<Arc<BstImageView>>,
 	sub_imgs: HashMap<SubImageID, SubImage>,
-	sto_leases: Vec<Vec<Arc<AtomicBool>>>,
 	con_sub_img: Vec<Vec<SubImageID>>,
 	alloc_cell_w: usize,
 	alloc: Vec<Vec<Option<SubImageID>>>,
@@ -803,7 +998,6 @@ impl AtlasImage {
 			active: None,
 			update: None,
 			sto_imgs: Vec::new(),
-			sto_leases: Vec::new(),
 			sub_imgs: HashMap::new(),
 			con_sub_img: Vec::new(),
 		}
@@ -818,9 +1012,7 @@ impl AtlasImage {
 			None => *self.active.as_ref()?,
 		};
 
-		let (tmp_img, abool) = self.sto_imgs[img_i].create_tmp();
-		self.sto_leases[img_i].push(abool);
-		Some(tmp_img)
+		Some(self.sto_imgs[img_i].create_tmp())
 	}
 
 	// TODO: Borrow cmd_buf
@@ -836,9 +1028,7 @@ impl AtlasImage {
 		let mut resize = false;
 
 		for (i, sto_img) in self.sto_imgs.iter().enumerate() {
-			self.sto_leases[i].retain(|v| v.load(atomic::Ordering::Relaxed));
-
-			if found_op.is_none() && self.sto_leases[i].is_empty() {
+			if found_op.is_none() && sto_img.temporary_views() == 0 {
 				if let VkImgDimensions::Dim2d {
 					width,
 					height,
@@ -874,7 +1064,7 @@ impl AtlasImage {
 						height: min_img_h,
 						array_layers: 1,
 					},
-					vulkano::format::Format::A8B8G8R8SrgbPack32,
+					self.basalt.formats_in_use().atlas,
 					VkImageUsage {
 						transfer_source: true,
 						transfer_destination: true,
@@ -882,9 +1072,11 @@ impl AtlasImage {
 						..VkImageUsage::none()
 					},
 					ImageCreateFlags::none(),
-					vec![self.basalt.compute_queue_ref().family()], /* TODO: Secondary
-					                                                 * graphics
-					                                                 * queue */
+					vec![self
+						.basalt
+						.secondary_graphics_queue_ref()
+						.unwrap_or_else(|| self.basalt.graphics_queue_ref())
+						.family()],
 				)
 				.unwrap(),
 			)
@@ -915,17 +1107,19 @@ impl AtlasImage {
 				)
 				.unwrap();
 
-				cmd_buf
-					.copy_buffer_to_image_dimensions(
-						r_buf,
-						image.clone(),
-						[cur_img_w, 0, 0],
-						[r_w, r_h, 0],
-						0,
-						1,
-						0,
-					)
-					.unwrap();
+				if r_w > 0 && r_h > 0 {
+					cmd_buf
+						.copy_buffer_to_image_dimensions(
+							r_buf,
+							image.clone(),
+							[cur_img_w, 0, 0],
+							[r_w, r_h, 1],
+							0,
+							1,
+							0,
+						)
+						.unwrap();
+				}
 
 				let b_w = min_img_w;
 				let b_h = min_img_h - cur_img_h;
@@ -943,17 +1137,19 @@ impl AtlasImage {
 				)
 				.unwrap();
 
-				cmd_buf
-					.copy_buffer_to_image_dimensions(
-						b_buf,
-						image.clone(),
-						[0, cur_img_h, 0],
-						[b_w, b_h, 0],
-						0,
-						1,
-						0,
-					)
-					.unwrap();
+				if b_w > 0 && b_h > 0 {
+					cmd_buf
+						.copy_buffer_to_image_dimensions(
+							b_buf,
+							image.clone(),
+							[0, cur_img_h, 0],
+							[b_w, b_h, 1],
+							0,
+							1,
+							0,
+						)
+						.unwrap();
+				}
 
 				cmd_buf
 					.copy_image(
@@ -970,8 +1166,8 @@ impl AtlasImage {
 					)
 					.unwrap();
 
+				self.sto_imgs[img_i].mark_stale();
 				self.sto_imgs[img_i] = image.clone();
-				self.sto_leases[img_i].clear();
 				found_op = Some((img_i, image));
 				cur_img_w = min_img_w;
 				cur_img_h = min_img_h;
@@ -979,7 +1175,6 @@ impl AtlasImage {
 				cmd_buf.clear_color_image(image.clone(), [0_32; 4].into()).unwrap();
 				self.sto_imgs.push(image.clone());
 				self.con_sub_img.push(Vec::new());
-				self.sto_leases.push(Vec::new());
 				found_op = Some((img_i, image));
 				self.update = Some(img_i);
 				cur_img_w = min_img_w;
@@ -990,33 +1185,62 @@ impl AtlasImage {
 		let (img_i, sto_img) = found_op.unwrap();
 		let mut upload_data = Vec::new();
 		let mut copy_cmds = Vec::new();
+		let mut copy_cmds_imt = Vec::new();
+		let mut copy_cmds_bst = Vec::new();
 
 		for (sub_img_id, sub_img) in &self.sub_imgs {
 			if !self.con_sub_img[img_i].contains(sub_img_id) {
-				if let ImageData::D8(sub_img_data) = &sub_img.img.data {
-					assert!(ImageType::SRGBA == sub_img.img.ty);
-					assert!(!sub_img_data.is_empty());
+				match &sub_img.img.data {
+					ImageData::D16(sub_img_data) => {
+						assert!(ImageType::LRGBA == sub_img.img.ty);
+						assert!(!sub_img_data.is_empty());
 
-					let s = upload_data.len();
-					upload_data.extend_from_slice(&sub_img_data);
+						let s = upload_data.len() as u64;
+						upload_data.extend_from_slice(&sub_img_data);
 
-					copy_cmds.push((
-						s,
-						upload_data.len(),
-						sub_img.coords.x,
-						sub_img.coords.y,
-						sub_img.coords.w,
-						sub_img.coords.h,
-					));
+						copy_cmds.push((
+							s,
+							upload_data.len() as u64,
+							sub_img.coords.x,
+							sub_img.coords.y,
+							sub_img.coords.w,
+							sub_img.coords.h,
+						));
 
-					self.con_sub_img[img_i].push(*sub_img_id);
-				} else {
-					unreachable!()
+						self.con_sub_img[img_i].push(*sub_img_id);
+					},
+					ImageData::Imt(view) => {
+						assert!(ImageType::Raw == sub_img.img.ty);
+
+						copy_cmds_imt.push((
+							view.clone(),
+							sub_img.coords.x,
+							sub_img.coords.y,
+							sub_img.coords.w,
+							sub_img.coords.h,
+						));
+
+						self.con_sub_img[img_i].push(*sub_img_id);
+					},
+					ImageData::Bst(view) => {
+						assert!(ImageType::Raw == sub_img.img.ty);
+
+						copy_cmds_bst.push((
+							view.clone(),
+							sub_img.coords.x,
+							sub_img.coords.y,
+							sub_img.coords.w,
+							sub_img.coords.h,
+						));
+
+						self.con_sub_img[img_i].push(*sub_img_id);
+					},
+					_ => unreachable!(),
 				}
 			}
 		}
 
-		if copy_cmds.is_empty() {
+		if copy_cmds.is_empty() && copy_cmds_imt.is_empty() && copy_cmds_bst.is_empty() {
 			self.update = None;
 			return (cmd_buf, false, cur_img_w, cur_img_h);
 		}
@@ -1038,7 +1262,7 @@ impl AtlasImage {
 					upload_buf.clone().into_buffer_slice().slice(s..e).unwrap(),
 					sto_img.clone(),
 					[x, y, 0],
-					[w, h, 0],
+					[w, h, 1],
 					0,
 					1,
 					0,
@@ -1046,6 +1270,78 @@ impl AtlasImage {
 				.unwrap();
 		}
 
+		for (v, x, y, w, h) in copy_cmds_imt {
+			if v.format() == sto_img.format() {
+				cmd_buf
+					.copy_image(
+						v,
+						[0, 0, 0],
+						0,
+						0,
+						sto_img.clone(),
+						[x as i32, y as i32, 0],
+						0,
+						0,
+						[w, h, 1],
+						1,
+					)
+					.unwrap();
+			} else {
+				cmd_buf
+					.blit_image(
+						v,
+						[0, 0, 0],
+						[w as i32, h as i32, 1],
+						0,
+						0,
+						sto_img.clone(),
+						[x as i32, y as i32, 0],
+						[x as i32 + w as i32, y as i32 + h as i32, 1],
+						0,
+						0,
+						1,
+						vulkano::sampler::Filter::Nearest,
+					)
+					.unwrap();
+			}
+		}
+
+		for (v, x, y, w, h) in copy_cmds_bst {
+			if v.format() == sto_img.format() {
+				cmd_buf
+					.copy_image(
+						v,
+						[0, 0, 0],
+						0,
+						0,
+						sto_img.clone(),
+						[x as i32, y as i32, 0],
+						0,
+						0,
+						[w, h, 1],
+						1,
+					)
+					.unwrap();
+			} else {
+				cmd_buf
+					.blit_image(
+						v,
+						[0, 0, 0],
+						[w as i32, h as i32, 1],
+						0,
+						0,
+						sto_img.clone(),
+						[x as i32, y as i32, 0],
+						[x as i32 + w as i32, y as i32 + h as i32, 1],
+						0,
+						0,
+						1,
+						vulkano::sampler::Filter::Nearest,
+					)
+					.unwrap();
+			}
+		}
+
 		(cmd_buf, true, cur_img_w, cur_img_h)
 	}
 
