@@ -1,14 +1,12 @@
 pub mod style;
 pub use self::style::{BinPosition, BinStyle, BinVert, Color, ImageEffect};
 
-use super::super::atlas;
-use super::interface::ItfVertInfo;
 use crate::atlas::{Image, ImageData, ImageDims, ImageType, SubImageCacheID};
 use crate::image_view::BstImageView;
 use crate::input::*;
 use crate::interface::hook::{BinHook, BinHookData, BinHookFn, BinHookID};
-use crate::interface::interface::scale_verts;
-use crate::{misc, Basalt};
+use crate::interface::{scale_verts, ItfVertInfo};
+use crate::{atlas, misc, Basalt};
 use arc_swap::ArcSwapAny;
 use ilmenite::*;
 use ordered_float::OrderedFloat;
@@ -174,6 +172,8 @@ pub struct PostUpdate {
 	pub z_index: i16,
 	pub pre_bound_min_y: f32,
 	pub pre_bound_max_y: f32,
+	pub extent: [u32; 2],
+	pub scale: f32,
 	text_state: Option<BinTextState>,
 }
 
@@ -200,6 +200,7 @@ struct BinTextStyle {
 	hori_align: ImtHoriAlign,
 }
 
+#[allow(dead_code)]
 #[derive(Clone, Default, Debug)]
 struct BinGlyphInfo {
 	min_x: f32,
@@ -1334,6 +1335,8 @@ impl Bin {
 			pre_bound_min_y: 0.0,
 			pre_bound_max_y: 0.0,
 			text_state: None,
+			extent: [win_size[0].trunc() as u32, win_size[1].trunc() as u32],
+			scale,
 		};
 
 		if update_stats {
@@ -1649,6 +1652,7 @@ impl Bin {
 						coords: (coords_x, coords_y),
 						color: back_color.as_tuple(),
 						ty,
+						tex_i: 0,
 					});
 				}
 			}
@@ -1660,36 +1664,42 @@ impl Bin {
 					coords: (0.0, 0.0),
 					color: border_color_t.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.tli[0], bps.tlo[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_t.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.tli[0], bps.tli[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_t.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.tri[0], bps.tro[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_t.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.tli[0], bps.tli[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_t.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.tri[0], bps.tri[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_t.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 			}
 			if border_color_b.a > 0.0 && border_size_b > 0.0 {
@@ -1699,36 +1709,42 @@ impl Bin {
 					coords: (0.0, 0.0),
 					color: border_color_b.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.bli[0], bps.bli[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_b.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.bli[0], bps.blo[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_b.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.bri[0], bps.bri[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_b.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.bli[0], bps.blo[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_b.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.bri[0], bps.bro[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_b.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 			}
 			if border_color_l.a > 0.0 && border_size_l > 0.0 {
@@ -1738,36 +1754,42 @@ impl Bin {
 					coords: (0.0, 0.0),
 					color: border_color_l.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.tlo[0], bps.tli[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_l.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.blo[0], bps.bli[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_l.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.tli[0], bps.tli[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_l.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.blo[0], bps.bli[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_l.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.bli[0], bps.bli[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_l.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 			}
 			if border_color_r.a > 0.0 && border_size_r > 0.0 {
@@ -1777,36 +1799,42 @@ impl Bin {
 					coords: (0.0, 0.0),
 					color: border_color_r.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.tri[0], bps.tri[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_r.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.bri[0], bps.bri[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_r.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.tro[0], bps.tri[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_r.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.bri[0], bps.bri[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_r.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.bro[0], bps.bri[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_r.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 			}
 			if border_color_t.a > 0.0
@@ -1820,18 +1848,21 @@ impl Bin {
 					coords: (0.0, 0.0),
 					color: border_color_l.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.tlo[0], bps.tli[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_l.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.tli[0], bps.tli[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_l.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				// Top Left Border Corner (Color of Top)
 				verts.push(ItfVertInfo {
@@ -1839,18 +1870,21 @@ impl Bin {
 					coords: (0.0, 0.0),
 					color: border_color_t.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.tlo[0], bps.tlo[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_t.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.tli[0], bps.tli[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_t.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 			}
 			if border_color_t.a > 0.0
@@ -1864,18 +1898,21 @@ impl Bin {
 					coords: (0.0, 0.0),
 					color: border_color_r.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.tri[0], bps.tri[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_r.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.tro[0], bps.tri[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_r.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				// Top Right Border Corner (Color of Top)
 				verts.push(ItfVertInfo {
@@ -1883,18 +1920,21 @@ impl Bin {
 					coords: (0.0, 0.0),
 					color: border_color_t.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.tri[0], bps.tro[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_t.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.tri[0], bps.tri[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_t.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 			}
 			if border_color_b.a > 0.0
@@ -1908,18 +1948,21 @@ impl Bin {
 					coords: (0.0, 0.0),
 					color: border_color_l.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.blo[0], bps.bli[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_l.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.blo[0], bps.blo[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_l.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				// Bottom Left Border Corner (Color of Bottom)
 				verts.push(ItfVertInfo {
@@ -1927,18 +1970,21 @@ impl Bin {
 					coords: (0.0, 0.0),
 					color: border_color_b.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.blo[0], bps.blo[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_b.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.bli[0], bps.blo[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_b.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 			}
 			if border_color_b.a > 0.0
@@ -1952,18 +1998,21 @@ impl Bin {
 					coords: (0.0, 0.0),
 					color: border_color_r.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.bri[0], bps.bri[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_r.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.bro[0], bps.bro[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_r.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				// Bottom Right Border Corner (Color of Bottom)
 				verts.push(ItfVertInfo {
@@ -1971,18 +2020,21 @@ impl Bin {
 					coords: (0.0, 0.0),
 					color: border_color_b.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.bri[0], bps.bro[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_b.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.bro[0], bps.bro[1], base_z),
 					coords: (0.0, 0.0),
 					color: border_color_b.as_tuple(),
 					ty: 0,
+					tex_i: 0,
 				});
 			}
 			if back_color.a > 0.0 || back_coords.img_id != 0 || back_img.is_some() {
@@ -1997,36 +2049,42 @@ impl Bin {
 					coords: back_coords.top_right(),
 					color: back_color.as_tuple(),
 					ty,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.tli[0], bps.tli[1], base_z),
 					coords: back_coords.top_left(),
 					color: back_color.as_tuple(),
 					ty,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.bli[0], bps.bli[1], base_z),
 					coords: back_coords.bottom_left(),
 					color: back_color.as_tuple(),
 					ty,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.tri[0], bps.tri[1], base_z),
 					coords: back_coords.top_right(),
 					color: back_color.as_tuple(),
 					ty,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.bli[0], bps.bli[1], base_z),
 					coords: back_coords.bottom_left(),
 					color: back_color.as_tuple(),
 					ty,
+					tex_i: 0,
 				});
 				verts.push(ItfVertInfo {
 					position: (bps.bri[0], bps.bri[1], base_z),
 					coords: back_coords.bottom_right(),
 					color: back_color.as_tuple(),
 					ty,
+					tex_i: 0,
 				});
 			}
 		}
@@ -2048,6 +2106,7 @@ impl Bin {
 				coords: (0.0, 0.0),
 				color: color.as_tuple(),
 				ty: 0,
+				tex_i: 0,
 			});
 		}
 
@@ -2067,8 +2126,8 @@ impl Bin {
 				let pad_b = style.pad_b.clone().unwrap_or(0.0);
 				let pad_l = style.pad_l.clone().unwrap_or(0.0);
 				let pad_r = style.pad_r.clone().unwrap_or(0.0);
-				let body_width = (bps.tri[0] - bps.tli[0] - pad_l - pad_r) * scale;
-				let body_height = (bps.bli[1] - bps.tli[1] - pad_t - pad_b) * scale;
+				let body_width = bps.tri[0] - bps.tli[0] - pad_l - pad_r;
+				let body_height = bps.bli[1] - bps.tli[1] - pad_t - pad_b;
 				let mut color = style.text_color.clone().unwrap_or(Color::srgb_hex("000000"));
 				color.a *= opacity;
 				let text_height = style.text_height.clone().unwrap_or(12.0);
@@ -2117,6 +2176,7 @@ impl Bin {
 								coords: vert.coords.clone(),
 								color: color.as_tuple(),
 								ty: 2,
+								tex_i: 0,
 							});
 						}
 					}
@@ -2126,10 +2186,10 @@ impl Bin {
 						ImtWeight::Normal,
 						text_height * scale,
 						Some(ImtShapeOpts {
-							body_width,
-							body_height,
-							text_height,
-							line_spacing,
+							body_width: body_width * scale,
+							body_height: body_height * scale,
+							text_height: text_height * scale,
+							line_spacing: line_spacing * scale,
 							text_wrap,
 							vert_align,
 							hori_align,
@@ -2160,7 +2220,7 @@ impl Bin {
 									glyph.family.clone(),
 									glyph.weight.clone(),
 									glyph.index,
-									OrderedFloat::from(text_height),
+									OrderedFloat::from(text_height * scale),
 								)
 							})
 							.collect(),
@@ -2178,35 +2238,50 @@ impl Bin {
 										glyph.family,
 										glyph.weight,
 										glyph.index,
-										OrderedFloat::from(text_height),
+										OrderedFloat::from(text_height * scale),
 									);
 
-									self.basalt
-										.atlas_ref()
-										.load_image(
-											cache_id,
-											Image::new(
-												ImageType::LRGBA,
-												ImageDims {
-													w: glyph.w,
-													h: glyph.h,
-												},
-												ImageData::D8(
-													glyph
-														.bitmap
-														.as_ref()
-														.unwrap()
-														.iter()
-														.map(|v| {
-															(*v * u8::max_value() as f32)
-																.round() as u8
-														})
-														.collect(),
-												),
-											)
-											.unwrap(),
-										)
-										.unwrap()
+									match glyph.bitmap.as_ref() {
+										Some(bitmap_data) =>
+											match bitmap_data {
+												ImtBitmapData::Empty => continue,
+												ImtBitmapData::LRGBA(image_data) =>
+													self.basalt
+														.atlas_ref()
+														.load_image(
+															cache_id,
+															Image::new(
+																ImageType::LRGBA,
+																ImageDims {
+																	w: glyph.w,
+																	h: glyph.h,
+																},
+																ImageData::D8(
+																	image_data
+																		.iter()
+																		.map(|v| {
+																			(*v * u8::max_value(
+																			) as f32)
+																				.round() as u8
+																		})
+																		.collect(),
+																),
+															)
+															.unwrap(),
+														)
+														.unwrap(),
+												ImtBitmapData::Image(view) =>
+													self.basalt
+														.atlas_ref()
+														.load_image(
+															cache_id,
+															Image::from_imt(view.clone())
+																.unwrap(),
+														)
+														.unwrap(),
+											},
+										None => continue,
+									}
 								},
 							}
 						};
@@ -2229,6 +2304,7 @@ impl Bin {
 							coords: (c_max_x, c_min_y),
 							color: color.as_tuple(),
 							ty: 2,
+							tex_i: 0,
 						});
 
 						verts.push(ItfVertInfo {
@@ -2236,6 +2312,7 @@ impl Bin {
 							coords: (c_min_x, c_min_y),
 							color: color.as_tuple(),
 							ty: 2,
+							tex_i: 0,
 						});
 
 						verts.push(ItfVertInfo {
@@ -2243,6 +2320,7 @@ impl Bin {
 							coords: (c_min_x, c_max_y),
 							color: color.as_tuple(),
 							ty: 2,
+							tex_i: 0,
 						});
 
 						verts.push(ItfVertInfo {
@@ -2250,6 +2328,7 @@ impl Bin {
 							coords: (c_max_x, c_min_y),
 							color: color.as_tuple(),
 							ty: 2,
+							tex_i: 0,
 						});
 
 						verts.push(ItfVertInfo {
@@ -2257,6 +2336,7 @@ impl Bin {
 							coords: (c_min_x, c_max_y),
 							color: color.as_tuple(),
 							ty: 2,
+							tex_i: 0,
 						});
 
 						verts.push(ItfVertInfo {
@@ -2264,6 +2344,7 @@ impl Bin {
 							coords: (c_max_x, c_max_y),
 							color: color.as_tuple(),
 							ty: 2,
+							tex_i: 0,
 						});
 
 						text_state.glyphs.push(BinGlyphInfo {
@@ -2462,7 +2543,7 @@ impl Bin {
 
 	pub fn force_update(&self) {
 		self.update.store(true, atomic::Ordering::SeqCst);
-		self.basalt.interface_ref().odb.unpark();
+		self.basalt.interface_ref().composer_ref().unpark();
 	}
 
 	pub fn update_children(&self) {
@@ -2472,6 +2553,7 @@ impl Bin {
 	fn update_children_priv(&self, update_self: bool) {
 		if update_self {
 			self.update.store(true, atomic::Ordering::SeqCst);
+			self.basalt.interface_ref().composer_ref().unpark();
 		}
 
 		for child in self.children().into_iter() {
@@ -2491,7 +2573,7 @@ impl Bin {
 		self.style.store(Arc::new(copy));
 		*self.initial.lock() = false;
 		self.update.store(true, atomic::Ordering::SeqCst);
-		self.basalt.interface_ref().odb.unpark();
+		self.basalt.interface_ref().composer_ref().unpark();
 	}
 
 	pub fn hidden(self: &Arc<Self>, to: Option<bool>) {
@@ -2508,7 +2590,7 @@ impl Bin {
 		});
 
 		self.update.store(true, atomic::Ordering::SeqCst);
-		self.basalt.interface_ref().odb.unpark();
+		self.basalt.interface_ref().composer_ref().unpark();
 	}
 
 	pub fn set_raw_img_yuv_422(
@@ -2527,7 +2609,7 @@ impl Bin {
 				array_layers: 1,
 			},
 			vulkano::image::MipmapsCount::One,
-			vulkano::format::Format::R8Unorm,
+			vulkano::format::Format::R8_UNORM,
 			self.basalt.transfer_queue(),
 		)
 		.unwrap();
@@ -2542,7 +2624,7 @@ impl Bin {
 		});
 
 		self.update.store(true, atomic::Ordering::SeqCst);
-		self.basalt.interface_ref().odb.unpark();
+		self.basalt.interface_ref().composer_ref().unpark();
 		Ok(())
 	}
 
@@ -2561,7 +2643,7 @@ impl Bin {
 					array_layers: 1,
 				},
 				vulkano::image::MipmapsCount::One,
-				vulkano::format::Format::R8G8B8A8Unorm,
+				vulkano::format::Format::R8G8B8A8_UNORM,
 				self.basalt.transfer_queue(),
 			)
 			.unwrap()
@@ -2575,28 +2657,10 @@ impl Bin {
 		});
 
 		self.update.store(true, atomic::Ordering::SeqCst);
-		self.basalt.interface_ref().odb.unpark();
+		self.basalt.interface_ref().composer_ref().unpark();
 		Ok(())
 	}
 
-	// pub fn set_raw_back_data(&self, width: u32, height: u32, data: Vec<u8>) -> Result<(),
-	// String> { self.basalt.atlas_ref().remove_raw(self.id);
-	//
-	// let coords = match self.basalt.atlas_ref().load_raw(self.id, data, width, height) {
-	// Ok(ok) => ok,
-	// Err(e) => return Err(e)
-	// };
-	//
-	// self.back_image.lock() = Some(ImageInfo {
-	// image: None,
-	// coords: coords
-	// });
-	//
-	// self.update.store(true, atomic::Ordering::SeqCst);
-	// self.basalt.interface_ref().odb.unpark();
-	// Ok(())
-	// }
-
 	pub fn remove_raw_back_img(&self) {
 		self.style_update(BinStyle {
 			back_image_raw: None,
@@ -2604,7 +2668,7 @@ impl Bin {
 		});
 
 		self.update.store(true, atomic::Ordering::SeqCst);
-		self.basalt.interface_ref().odb.unpark();
+		self.basalt.interface_ref().composer_ref().unpark();
 	}
 }
 
